"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createEffect = void 0;
const rxjs_1 = require("rxjs");
const effectController_1 = require("./effectController");
const DEFAULT_MERGE_MAP_PIPELINE = (eventProject) => (0, rxjs_1.mergeMap)(eventProject);
/**
 * Creates `Effect` from the provided handler.
 *
 * @example
 * ```ts
 * const sumEffect = createEffect<{a: number, b: number}, number>((event) => {
 *   return a + b;
 * });
 * ```
 */
function createEffect(handler, options) {
    var _a;
    const pipeline = (_a = options === null || options === void 0 ? void 0 : options.pipeline) !== null && _a !== void 0 ? _a : DEFAULT_MERGE_MAP_PIPELINE;
    const event$ = new rxjs_1.Subject();
    const controller = (0, effectController_1.createEffectController)();
    const subscriptions = new rxjs_1.Subscription(() => {
        event$.complete();
        controller.destroy();
    });
    const eventProject = (event) => {
        return (0, rxjs_1.defer)(() => {
            controller.start();
            const result = handler(event);
            return result instanceof rxjs_1.Observable || result instanceof Promise
                ? result
                : (0, rxjs_1.of)(result);
        }).pipe((0, rxjs_1.tap)({
            next: (result) => {
                controller.next({ event, result });
            },
            complete: () => {
                controller.complete();
            },
            error: (error) => {
                controller.error({ origin: 'handler', event, error });
            },
        }));
    };
    subscriptions.add(event$.pipe(pipeline(eventProject), (0, rxjs_1.retry)()).subscribe());
    return {
        ...controller.state,
        handle(source) {
            const observable = getSourceObservable(source);
            const subscription = observable.subscribe({
                next: (event) => event$.next(event),
                error: (error) => controller.error({ origin: 'source', error }),
            });
            subscriptions.add(subscription);
            return subscription;
        },
        destroy: () => {
            subscriptions.unsubscribe();
        },
    };
}
exports.createEffect = createEffect;
function getSourceObservable(source) {
    const type = typeof source;
    if (type === 'function' && 'event$' in source) {
        return source.event$;
    }
    if (type === 'object' && 'value$' in source) {
        return source.value$;
    }
    if (source instanceof rxjs_1.Observable) {
        return source;
    }
    throw new TypeError('Unexpected source type');
}
//# sourceMappingURL=effect.js.map