"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.withStoreUpdates = exports.createStoreUpdates = exports.createStore = exports.pipeStateMutations = exports.declareStateUpdates = void 0;
const rxjs_1 = require("rxjs");
const queryMappers_1 = require("./queryMappers");
const storeEvents_1 = require("./storeEvents");
const storeMetadata_1 = require("./storeMetadata");
const utils_1 = require("./utils");
let STORE_SEQ_NUMBER = 0;
function declareStateUpdates(stateExample, updates) {
    if (updates) {
        return updates;
    }
    return (updates) => updates;
}
exports.declareStateUpdates = declareStateUpdates;
/**
 * Returns a mutation which applies all provided mutations for a state.
 *
 * You can use this helper to apply multiple changes at the same time.
 */
function pipeStateMutations(mutations) {
    return (state) => mutations.reduce((nextState, mutation) => mutation(nextState), state);
}
exports.pipeStateMutations = pipeStateMutations;
/**
 * Creates the state store.
 *
 * @param initialState Initial state
 * @param options Parameters for the store
 */
function createStore(initialState, options) {
    var _a;
    const stateComparator = (_a = options === null || options === void 0 ? void 0 : options.comparator) !== null && _a !== void 0 ? _a : utils_1.DEFAULT_COMPARATOR;
    const store$ = new rxjs_1.BehaviorSubject(initialState);
    const state$ = store$.asObservable();
    let isUpdating = false;
    let pendingMutations;
    const store = {
        id: ++STORE_SEQ_NUMBER,
        name: options === null || options === void 0 ? void 0 : options.name,
        value$: state$,
        get() {
            return store$.value;
        },
        select(selector, options) {
            return this.query(selector, options).value$;
        },
        query(selector, options) {
            return (0, queryMappers_1.mapQuery)(this, selector, options);
        },
        asQuery: () => store,
        set(nextState) {
            apply([() => nextState]);
        },
        update,
        destroy() {
            var _a;
            store$.complete();
            storeEvents_1.STORE_EVENT_BUS.next({ type: 'destroyed', store });
            (_a = options === null || options === void 0 ? void 0 : options.onDestroy) === null || _a === void 0 ? void 0 : _a.call(options);
        },
    };
    function update(mutation) {
        if ((0, utils_1.isReadonlyArray)(mutation)) {
            apply(mutation);
        }
        else {
            apply([mutation]);
        }
    }
    function apply(mutations) {
        if (isUpdating) {
            pendingMutations = (pendingMutations !== null && pendingMutations !== void 0 ? pendingMutations : []).concat(mutations);
            return;
        }
        const prevState = store$.value;
        let nextState = prevState;
        for (let i = 0; i < mutations.length; i++) {
            const mutation = mutations[i];
            if (mutation) {
                const stateBeforeMutation = nextState;
                nextState = mutation(nextState);
                storeEvents_1.STORE_EVENT_BUS.next({
                    type: 'mutation',
                    store,
                    mutation,
                    nextState,
                    prevState: stateBeforeMutation,
                });
            }
        }
        if (stateComparator(prevState, nextState)) {
            return;
        }
        isUpdating = true;
        store$.next(nextState);
        storeEvents_1.STORE_EVENT_BUS.next({
            type: 'updated',
            store,
            nextState,
            prevState,
        });
        isUpdating = false;
        if (pendingMutations === null || pendingMutations === void 0 ? void 0 : pendingMutations.length) {
            const mutationsToApply = pendingMutations;
            pendingMutations = [];
            apply(mutationsToApply);
        }
    }
    if (options === null || options === void 0 ? void 0 : options.internal) {
        (0, storeMetadata_1.setInternalStoreFlag)(store);
    }
    storeEvents_1.STORE_EVENT_BUS.next({ type: 'created', store });
    return store;
}
exports.createStore = createStore;
/** Creates StateUpdates for updating the store by provided state mutations */
function createStoreUpdates(storeUpdate, stateUpdates) {
    const updates = {};
    Object.entries(stateUpdates).forEach(([key, mutationFactory]) => {
        updates[key] = (...args) => {
            const mutation = mutationFactory(...args);
            (0, storeMetadata_1.setStateMutationName)(mutation, key);
            storeUpdate(mutation);
        };
    });
    return updates;
}
exports.createStoreUpdates = createStoreUpdates;
/** Creates a proxy for the store with "updates" to change a state by provided mutations */
function withStoreUpdates(store, updates) {
    const storeUpdates = createStoreUpdates(store.update, updates);
    return { ...store, updates: storeUpdates };
}
exports.withStoreUpdates = withStoreUpdates;
//# sourceMappingURL=store.js.map